#include "global.h"
#include <async.h>
#include <KVM.h>
#include <SerialMgr.h>

UInt serref = 0;
static SerSettingsType serial_settings;
UInt ticksToWait = -1;
int firstTime = TRUE;

/*=======================================================================
 * Forward declarations
 *=======================================================================*/

void prim_realOpen(INSTANCE instance, char **exception, int baud, 
		   int flags, int ctsTimeout, int receiveTimeout);
int prim_readBytes(INSTANCE instance, char **exception, 
		   char *buffer, int offset, int length);
void prim_writeBytes(INSTANCE instance, char **exception, 
		     char *buffer, int offset, int length);


void Java_com_sun_cldc_io_palm_comm_Protocol_prim_1realOpen(void);
void Java_com_sun_cldc_io_palm_comm_Protocol_prim_1close(void);
void Java_com_sun_cldc_io_palm_comm_Protocol_prim_1readBytes(void);
void Java_com_sun_cldc_io_palm_comm_Protocol_prim_1writeBytes(void);

/*
 * prim_readBytes
 */
void prim_realOpen(INSTANCE instance, char **exception, int baud, int flags, int ctsTimeout, int receiveTimeout) {
    Err errout;
    ULong starttime;
    ULong bytecount;


    if(firstTime) {
        if ((errout = SysLibFind("Serial Library", &serref)) != 0) {
            *exception = "java/io/IOException";
            return;
        }
        firstTime = FALSE;
    }

    serial_settings.baudRate   = baud;
    serial_settings.ctsTimeout = serDefaultCTSTimeout;
    serial_settings.flags =
        serSettingsFlagBitsPerChar8 |
        serSettingsFlagStopBits1 |
        serSettingsFlagCTSAutoM |
        serSettingsFlagRTSAutoM;

    if(flags != -1) {
        serial_settings.flags = flags;
    }

    if(ctsTimeout != -1) {
        serial_settings.ctsTimeout = ctsTimeout;
    }

    if ((errout = SerOpen(serref, 0, serial_settings.baudRate)) != 0 &&
	errout != serErrAlreadyOpen) {
        *exception = "java/io/IOException";
        return; // perhaps net is connected
    }

    if ((errout = SerSetSettings(serref, &serial_settings)) != 0) {
        SerClearErr(serref);
        *exception = "java/io/IOException";
        return;
    }

    starttime = TimGetSeconds();
    while (starttime + 2 > TimGetSeconds()){
        /* Just spin */
    }

    if (SerReceiveCheck(serref, &bytecount) != 0) {
        SerClearErr(serref);
    }

    if (bytecount > 0) {
        SerReceiveFlush(serref, 1);
    }

    ticksToWait = receiveTimeout;

    return;
}

/*
 * native void prim_realOpen(int baud, int flags, int ctsTimeout, int receiveTimeout) throws IOException;
 */
ASYNC_FUNCTION(Java_com_sun_cldc_io_palm_comm_Protocol_prim_1realOpen) {
    long            receiveTimeout = ASYNC_POP_STACK();
    long            ctsTimeout     = ASYNC_POP_STACK();
    long            flags          = ASYNC_POP_STACK();
    long            baud           = ASYNC_POP_STACK();
    INSTANCE        instance       = ASYNC_POP_STACK_AS_TYPE(INSTANCE);
    char           *exception      = NULL;

    prim_realOpen(instance, &exception, baud, flags, ctsTimeout, receiveTimeout);

    if(exception) {
        ASYNC_EXCEPTION(exception);
    }
    ASYNC_RESUME_THREAD();
}



/*
 * native void prim_close() throws IOException;
 */
ASYNC_FUNCTION(Java_com_sun_cldc_io_palm_comm_Protocol_prim_1close) {
    INSTANCE        instance  = ASYNC_POP_STACK_AS_TYPE(INSTANCE);

    SerClose(serref);

    ASYNC_RESUME_THREAD();
}




/*
 * prim_readBytes
 */
int prim_readBytes(INSTANCE instance, char **exception, char *buffer, int offset, int length) {

    Long bytecount = 0;
    ULong readcount, countread;
    Err error = 0;
    int i;

    for (i = 0; i < 5; i++) {
        if ((error =
            SerReceiveCheck(serref, (ULong *)&bytecount)) != 0) {
            SerClearErr(serref);
        } else {
            break;
        }
    }

    if (bytecount <= 0)
        return 0;

    length = SerReceive(serref, buffer+offset, length, ticksToWait, &error);

#ifdef SERIALDEBUG
    fprintf(stderr, "prim_readBytes res = %d error = %d\n", length, error);
#endif

    if (error){
        SerClearErr(serref);
        *exception = "java/io/IOException";
    }

    return length;
}

/*
 * native int prim_readBytes(byte b[], int off, int len) throws IOException;
 */
ASYNC_FUNCTION(Java_com_sun_cldc_io_palm_comm_Protocol_prim_1readBytes) {
    long            length = ASYNC_POP_STACK();
    long            offset = ASYNC_POP_STACK();
    BYTEARRAY       buffer = ASYNC_POP_STACK_AS_TYPE(BYTEARRAY);
    INSTANCE        instance = ASYNC_POP_STACK_AS_TYPE(INSTANCE);
    char           *exception = NULL;

    int v = prim_readBytes(instance, &exception, (buffer->bdata), offset, length);

    if(exception) {
        ASYNC_EXCEPTION(exception);
    }
    ASYNC_PUSH_STACK(v);
    ASYNC_RESUME_THREAD();
}






/*
 * prim_writeBytes
 */
void prim_writeBytes(INSTANCE instance, char **exception, char *buffer, int offset, int length) {

    Err error = 0;

    SerSend(serref, buffer+offset, length, &error);
    if (error) {
        *exception = "java/io/IOException";
    } else {
        SerSendWait(serref, -1);
    }

#ifdef SERIALDEBUG
    fprintf(stderr, "prim_writeBytes res = %d error = %d\n", length, error);
#endif
}

/*
 * native int prim_writeBytes(byte b[], int off, int len) throws IOException;
 */
ASYNC_FUNCTION(Java_com_sun_cldc_io_palm_comm_Protocol_prim_1writeBytes) {
    long            length = ASYNC_POP_STACK();
    long            offset = ASYNC_POP_STACK();
    BYTEARRAY       buffer = ASYNC_POP_STACK_AS_TYPE(BYTEARRAY);
    INSTANCE        instance = ASYNC_POP_STACK_AS_TYPE(INSTANCE);
    char           *exception = NULL;

    prim_writeBytes(instance, &exception, (buffer->bdata), offset, length);

    if(exception) {
        ASYNC_EXCEPTION(exception);
    }
    ASYNC_RESUME_THREAD();
}
